import loDefaultsDeep from 'lodash/defaultsDeep.js';
import loIsPlainObject from 'lodash/isPlainObject.js';
import loIsFunction from 'lodash/isFunction.js';
import loIsString from 'lodash/isString.js';
import { BUSINESS_OK } from '../common/keys.js';
import { isBlob } from '../common/types.js';
import { ElMessage } from 'element-plus';

export class AxiosRequest {

    constructor(serviceEngine, serviceConfig = null, config = null) {
        this.serviceEngine = serviceEngine;
        this.service = serviceConfig ? this.serviceEngine.create(serviceConfig) : this.serviceEngine.create();
        this.method = String();
        this.path = String();
        this.params = Object();
        this.queries = Object();
        this.body = Object();
        this.defaultConfig = config || Object();
        this.config = Object();
        this.authType = String();
        this.authToken = String();
        this.AUTH_TYPE_BEARER = String('Bearer');
    }

    init() {
        const self = this;

        self.defaultConfig = {

            url: '',

            method: 'get', // 默认值

            baseURL: '',

            transformResponse: [
                (response) => {
                    try {
                        if (isBlob(response)) {
                            return {
                                code: BUSINESS_OK,
                                data: response,
                            };
                        } else {
                            const businessData = JSON.parse(response);
                            let code = Number();
                            let data = Object();
                            if (
                                businessData.code === null
                                || businessData.code === undefined
                            ) {
                                code = NaN;
                                data = businessData;
                            } else if (businessData.code === BUSINESS_OK) {
                                code = BUSINESS_OK;
                                data = businessData.data;
                            } else {
                                code = businessData.code;
                                data = businessData.message;
                            }
                            return { code, data };
                        }
                    } catch (error) {
                        return { code: NaN, data: error };
                    }
                }
            ],

            headers: {},

            params: {},

            data: {},

            withCredentials: false, // 默认值

            responseType: 'json', // 默认值

            validateStatus: function (status) {
                return status >= 200 && status < 300; // 默认值
            },

            cancelToken: new self.serviceEngine.CancelToken(function (cancel) { }),

            // 构造器传进来的初始配置集合
            ...self.defaultConfig,

        }

        self.service.interceptors.request.use(
            config => config,
            error => {
                return Promise.reject(error);
            }
        );

        self.service.interceptors.response.use(
            response => {
                const businessData = response.data;
                if (isBlob(businessData.data)) {
                    const tagA = document.createElement('a');
                    const url = window.URL.createObjectURL(businessData.data);
                    tagA.href = url;
                    tagA.download = response.headers['content-disposition']
                        ?.split?.(';')
                        .find(row => row?.includes?.('filename=') ?? false)
                        ?.split?.('=')[1]
                        ?? 'no-filename-file';
                    const tagBody = document.getElementsByTagName('body')[0];
                    tagBody.appendChild(tagA);
                    tagA.click();
                    tagBody.removeChild(tagA);
                    window.URL.revokeObjectURL(url);
                }
                return response;
            },
            error => {
                return Promise.reject(error);
            }
        );

        return self;
    }

    setAuthType(type) {
        this.authType = type || '';
        return this;
    }

    setAuthToken(token) {
        this.authToken = token;
        return this;
    }

    setMethod(method) {
        this.method = method;
        return this;
    }

    setPath(path) {
        this.path = path;
        return this;
    }

    setParams(params) {
        const self = this;
        self.params = params;
        self.path = self.path
            .replace(
                /\{([^\{\}]+)\}/g,
                (arg1, $1, arg2, arg3) => self.params[$1]);
        return self;
    }

    setQueries(queries) {
        this.queries = queries;
        return this;
    }

    setBody(body) {
        this.body = body;
        return this;
    }

    setConfig(propName, propValue) {
        if (loIsPlainObject(propValue)) {
            this.config[propName] = loDefaultsDeep(propValue, this.config[propName]);
        } else {
            this.config[propName] = propValue;
        }

        return this;
    }

    reset() {
        this.method = '';
        this.path = '';
        this.params = {};
        this.queries = {};
        this.body = {};
        this.config = {};
        return void 0;
    }

    auth() {
        if (this.authType === this.AUTH_TYPE_BEARER) {
            const authType = this.authType;
            let authToken = '';
            if (loIsFunction(this.authToken)) {
                authToken = this.authToken();
            } else if (loIsString(this.authToken)) {
                authToken = this.authToken;
            }
            this.setConfig('headers', {
                Authorization: `${authType} ${authToken}`,
            });
        }
        return this;
    }

    send() {
        const self = this;
        return new Promise((resolve, reject) => {
            self.auth()
                .setConfig('method', self.method)
                .setConfig('url', self.path)
                .setConfig('params', self.queries)
                .setConfig('data', self.body);
            const config = {

                ...self.defaultConfig,

                ...self.config,

            };
            self
                .service(config)
                .then(res => {
                    return res;
                })
                .then(
                    (
                        {
                            data: { code, data }
                        }
                    ) => {
                        if (code === BUSINESS_OK) {
                            resolve(data);
                        } else {
                            throw {
                                response: {
                                    data: { code, data },
                                },
                            };
                        }
                    })
                .catch(
                    (
                        {
                            response: {
                                data: { code, data },
                            },
                        }
                    ) => {
                        const SYSTEM_ERROR = '系统错误';
                        const BUSINESS_EXCEPTION = '业务异常';
                        if (String(code) === 'NaN') {
                            ElMessage.error(`${SYSTEM_ERROR}，请联系运维人员处理，详情可查看浏览器控制台`);
                            console.error(`${SYSTEM_ERROR}:`, data);
                        } else {
                            ElMessage.warning(`${BUSINESS_EXCEPTION}，请稍后重试，或联系运维人员处理，详情可查看浏览器控制台`);
                            console.error(`${BUSINESS_EXCEPTION}:`, data);
                        }
                        reject({ code, data });
                    });

            self.reset();
        });
    }

};
AxiosRequest.NO_CONFIG = null;
