const writeFile = require('node:fs').writeFile;
const opendir = require('node:fs').opendir;
const mkdir = require('node:fs').mkdir;
const argv = require('node:process').argv;
const axios = require('axios');
const jsonFormat = require('json-format');

function fetchAPIDoc() {
    const jsonIndentConfig = {
        type: 'space',
        size: 4,
    };

    const eol = '\n';
    const indent = '    ';

    function indentToLevel(level = 1) {
        let t = '';
        while (t.length < level * 4) t += indent;
        return t;
    }

    const snippetsDir = './.vscode';

    const lastBrace = /(\})$/m;
    function lastBraceIndent(level) {
        const t = indentToLevel(level);
        return (arg1, $1, arg2, arg3) => `${t}${$1}`;
    }

    const argPatternAPIDocServer = /^--api-doc-server=(.+)$/;
    const argPatternAPIDocAlias = /^--api-doc-alias=(.+)$/;
    let apiDocServer = '';
    let apiDocAlias = '';
    argv.forEach(arg => {
        if (argPatternAPIDocServer.test(arg)) {
            apiDocServer = arg.replace(
                argPatternAPIDocServer,
                (arg1, $1, arg2, arg3) => $1
            );
        }

        if (argPatternAPIDocAlias.test(arg)) {
            apiDocAlias = arg.replace(
                argPatternAPIDocAlias,
                (arg1, $1, arg2, arg3) => $1
            );
        }
    });

    axios
        .get(apiDocServer)
        .then(response => {
            return response.data;
        })
        .then(({ components, paths, servers, info }) => {
            const scope = 'javascript';

            const httpResponseStatusOK = ['200', '201', '202', '203', '204', '205', '206', '207', '208'];
            const httpResponseAnyContent = '*/*';

            const reference = '$ref';

            const DOWNLOAD_KEYWORD = '下载';

            const server = servers.pop();
            const urlProxyFlag = server.url.replace(/.+\/api(\/.*)/, (arg1, $1, arg2, arg3) => $1);
            const serverDescription = info.title;

            function getPropInfo(prop) {
                let type = '';
                let description = '';
                let required = null;
                let defaultValue = null;
                let enums = null;
                switch (prop.type) {
                    case 'integer':
                    case 'number':
                        type = 'Number';
                        description = prop.description;
                        required = prop.required;
                        defaultValue = prop.default;
                        enums = prop.enum;
                        break;
                    case 'string':
                        type = 'String';
                        description = prop.description;
                        required = prop.required;
                        defaultValue = prop.default;
                        enums = prop.enum
                        break;
                    case 'boolean':
                        type = 'Boolean';
                        description = prop.description;
                        required = prop.required;
                        defaultValue = prop.default;
                        enums = prop.enum
                        break;
                    case 'array':
                        const baseTypeMap = {
                            integer: 'Number',
                            string: 'String',
                            boolean: 'Boolean',
                        };
                        const itemType = resolveEntity(prop);
                        const baseType = baseTypeMap[itemType];
                        type = `Array<${baseType || itemType}>`;
                        description = prop.description;
                        break;

                    default:
                        type = resolveEntity(prop);
                        if (type === 'plain-object') {
                            description = '任意内容对象，可放额外信息';
                        } else {
                            description = components.schemas[type].description;
                        }
                }

                return { type, description, required, defaultValue, enums };
            }
            function getNote(prop) {
                const { type, required = null, description, defaultValue = null, enums = null } = getPropInfo(prop);
                const typeNote = type || '未知类型';
                const requiredNote = required === null ? '' : `, ${required ? '必填' : '非必填'}`;
                const descriptionNote = description ? `, ${description}` : '';
                const defaultValueNote = defaultValue !== null ? `, 默认值: ${typeof defaultValue === 'string' ? `"${defaultValue}"` : defaultValue}` : '';
                const enumsNote = enums !== null ? `, 枚举值: [${enums.join(' | ')}]` : '';
                const note = `, // ${typeNote}${requiredNote}${descriptionNote}${defaultValueNote}${enumsNote}`;
                return note;
            }
            function resolveEntity(target, level) {
                const t = indentToLevel(level);
                const resolvedData = [];
                if (target.properties) {
                    resolvedData.push('{');
                    for (const [prop, propInfo] of Object.entries(target.properties)) {
                        resolvedData.push(
                            `${t}${prop}${getNote(propInfo)}`
                        );
                    }
                    resolvedData.push('}');
                    return resolvedData.join(eol);
                } else if (target[reference]) {
                    const schemaName = target[reference].split('/').pop();
                    return schemaName;
                } else if (target.type === 'array') {
                    const items = target.items;
                    if (items.type) {
                        return items.type;
                    } else if (items[reference]) {
                        const schemaName = items[reference].split('/').pop();
                        return schemaName;
                    }
                } else if (target.type === 'object') {
                    return 'plain-object';
                }
            }

            function makeEntitiesSnippets(entityName, entityJSON) {
                const snippets = this;
                const description = entityName;
                const prefix = entityName;
                const body = entityJSON.split(eol);
                snippets[description] = {
                    scope,
                    prefix,
                    body,
                    description,
                };
            }

            const entities = {};
            const entitiesSnippets = {};
            for (const [entityName, entityInfo] of Object.entries(components.schemas)) {
                let entityJSON = '';
                if (entityInfo.type === 'object' && entityInfo.oneOf) {
                    const genericCommentJSON = [
                        '/*',
                        ' * 根据业务场景不同, 将是这些枚举出的类型其中之一:',
                        ` * [${entityInfo.oneOf.map(one => one[reference].split('/').pop()).join(' | ')}]`,
                        ' */',
                    ].join(eol);
                    if (!entityInfo.properties) {
                        entityJSON = genericCommentJSON;
                        makeEntitiesSnippets.call(entitiesSnippets, entityName, entityJSON);
                    } else if (entityInfo.properties) {
                        let entityRows = null;
                        let genericCommentRows = null;
                        let indentedGenericCommentRows = null;
                        let tempToken = null;
                        let t = null;

                        const extrasCommentRow = ' * 额外地, 将解构下面的数据结构加入到当前结构中:';
                        genericCommentRows = genericCommentJSON.split(eol);
                        tempToken = genericCommentRows.shift();
                        genericCommentRows.unshift(extrasCommentRow);
                        genericCommentRows.unshift(tempToken);

                        entityJSON = resolveEntity(entityInfo, 2);
                        entityRows = entityJSON.split(eol);
                        t = indentToLevel(2);
                        indentedGenericCommentRows = genericCommentRows.map(row => `${t}${row}`);
                        tempToken = entityRows.pop();
                        entityRows.push(...indentedGenericCommentRows);
                        entityRows.push(tempToken);
                        entityJSON = entityRows.join(eol);
                        entities[entityName] = entityJSON;

                        entityJSON = resolveEntity(entityInfo);
                        entityRows = entityJSON.split(eol);
                        t = indentToLevel();
                        indentedGenericCommentRows = genericCommentRows.map(row => `${t}${row}`);
                        tempToken = entityRows.pop();
                        entityRows.push(...indentedGenericCommentRows);
                        entityRows.push(tempToken);
                        entityJSON = entityRows.join(eol);
                        makeEntitiesSnippets.call(entitiesSnippets, entityName, entityJSON);
                    }
                } else if (entityInfo.type === 'string') {
                    entityJSON = [
                        '/*',
                        ' * 根据业务场景不同, 将是这些枚举出的类型其中之一:',
                        ` * [${entityInfo.enum.join(' | ')}]`,
                        ' * 对于上述这些类型, 描述如下:',
                        ` * ${entityInfo.description}`,
                        ` * 默认为枚举值: ${entityInfo.default}`,
                        ' */',
                    ].join(eol);
                    makeEntitiesSnippets.call(entitiesSnippets, entityName, entityJSON);
                } else {
                    entityJSON = resolveEntity(entityInfo, 2);
                    entities[entityName] = entityJSON;

                    entityJSON = resolveEntity(entityInfo);
                    makeEntitiesSnippets.call(entitiesSnippets, entityName, entityJSON);
                }
            }

            const apiPathsSnippets = {};
            for (const [path, pathInfo] of Object.entries(paths)) {
                for (const [method, methodInfo] of Object.entries(pathInfo)) {
                    if (methodInfo.deprecated === true) continue;

                    const needDownload = methodInfo.summary?.includes?.(DOWNLOAD_KEYWORD) ?? false;

                    const parameters = methodInfo.parameters;
                    let hasParamInPath = false;
                    let paramsJSON = '';
                    let hasParamInQuery = false;
                    let queriesJSON = '';
                    if (parameters) {
                        const params = parameters.filter(param => param.in === 'path');
                        hasParamInPath = params.length > 0;
                        if (hasParamInPath) {
                            const paramMap = {
                                properties: {},
                            };
                            params.forEach(param => {
                                if (param.schema[reference]) {
                                    paramMap.properties[param.name] = param.schema;
                                } else {
                                    paramMap.properties[param.name] = {
                                        type: param.schema.type,
                                        description: param.description,
                                        required: param.required,
                                        default: param.schema.default,
                                        enum: param.schema.enum,
                                    };
                                }
                            });
                            paramsJSON = resolveEntity(paramMap, 2)
                                .replace(lastBrace, lastBraceIndent());
                        }

                        const queries = parameters.filter(param => param.in === 'query');
                        hasParamInQuery = queries.length > 0;
                        if (hasParamInQuery) {
                            const queryMap = {
                                properties: {},
                            };
                            queries.forEach(query => {
                                if (query.schema[reference]) {
                                    queryMap.properties[query.name] = query.schema;
                                } else {
                                    queryMap.properties[query.name] = {
                                        type: query.schema.type,
                                        description: query.description,
                                        required: query.required,
                                        default: query.schema.default,
                                        enum: query.schema.enum,
                                    };
                                }
                            });
                            queriesJSON = resolveEntity(queryMap, 2)
                                .replace(lastBrace, lastBraceIndent());
                        }
                    }

                    const requestBody = methodInfo.requestBody;
                    let hasBody = false;
                    let bodyJSON = '';
                    if (requestBody && requestBody.content['application/json']) {
                        const entity = requestBody.content['application/json'].schema;
                        let entityName = '';
                        hasBody = true;
                        if (entity.type === 'array') {
                            const items = entity.items;
                            const t = indentToLevel(2);
                            if (items.type) {
                                const { type } = getPropInfo(items);
                                bodyJSON = [
                                    '[',
                                    `${t}${type},`,
                                    ']',
                                ]
                                    .join(eol)
                                    .replace(lastBrace, lastBraceIndent());
                            } else if (items[reference]) {
                                entityName = items[reference].split('/').pop();
                                bodyJSON = [
                                    '[',
                                    `${t}${entities[entityName]},`,
                                    ']',
                                ]
                                    .join(eol)
                                    .replace(lastBrace, lastBraceIndent());
                            }
                        } else if (entity[reference]) {
                            entityName = entity[reference].split('/').pop();
                            bodyJSON = entities[entityName]
                                .replace(lastBrace, lastBraceIndent());
                        }
                    }

                    let responseBusinessNote = '';
                    const propValueIndex = 1;
                    const httpResponseStatusOKInfo = Object
                        .entries(methodInfo.responses)
                        .find((
                            [propName, propValue]
                        ) =>
                            httpResponseStatusOK.includes(propName) && propValue
                        )[propValueIndex];
                    if (httpResponseStatusOKInfo?.content?.[httpResponseAnyContent]?.schema?.properties?.data ?? false) {
                        const httpResponseContent = httpResponseStatusOKInfo.content;
                        const httpResponseBusinessData = httpResponseContent[httpResponseAnyContent].schema.properties.data;
                        responseBusinessNote = getNote(httpResponseBusinessData).substring(1);
                    } else {
                        responseBusinessNote = ' // Null, 业务操作成功没有返回的数据';
                    }

                    const prefix = path;
                    const description = [
                        method.toUpperCase(),
                        serverDescription,
                        methodInfo.summary,
                        prefix,
                    ].join(`,${eol}`);
                    const body = [
                        `req`,
                        `${indentToLevel()}.setMethod('${method}')`,
                        `${indentToLevel()}.setPath('${urlProxyFlag}${path}')`,
                    ];
                    needDownload && body.push(`${indentToLevel()}.setConfig('responseType', 'blob')`);
                    hasParamInPath && body.push(`${indentToLevel()}.setParams(${paramsJSON})`);
                    hasParamInQuery && body.push(`${indentToLevel()}.setQueries(${queriesJSON})`);
                    hasBody && body.push(`${indentToLevel()}.setBody(${bodyJSON})`);
                    body.push(
                        `${indentToLevel()}.send()`,
                        `${indentToLevel()}.then((`,
                        `${indentToLevel(2)}data${responseBusinessNote}`,
                        `${indentToLevel()}) => {`,
                        `${indentToLevel(2)}// 业务数据正常返回时的处理`,
                        `${indentToLevel()}})`,
                        `${indentToLevel()}.catch((rejectInfo) => {`,
                        `${indentToLevel(2)}/*`,
                        `${indentToLevel(2)} * 一般地, rejectInfo 结构是 { code, data }, 其中:`,
                        `${indentToLevel(2)} * rejectInfo.code 是业务异常码，或者，在发生系统错误时是 NaN;`,
                        `${indentToLevel(2)} * rejectInfo.data 是业务异常信息，或者，是系统错误信息;`,
                        `${indentToLevel(2)} * 特别地, 如果出现代码错误 rejectInfo 数据结构将是不可预测的, 此时默认会将其输出到控制台`,
                        `${indentToLevel(2)} */`,
                        `${indentToLevel(2)}const { code, data } = rejectInfo`,
                        `${indentToLevel(2)}if (code === null || code === undefined) {`,
                        `${indentToLevel(3)}console.error(rejectInfo);`,
                        `${indentToLevel(2)}} else if (code === NaN) {`,
                        `${indentToLevel(3)}// 请在这里处理系统错误情况`,
                        `${indentToLevel(2)}} else {`,
                        `${indentToLevel(3)}// 请在这里处理业务异常情况`,
                        `${indentToLevel(2)}}`,
                        `${indentToLevel()}})`,
                        `${indentToLevel()}.then((data) => {`,
                        `${indentToLevel(2)}/*`,
                        `${indentToLevel(2)} * finally-then, 可选, 是处理完所有 resolved 数据, 或者, rejected 数据, 最后无论什么情况都需要统一处理的逻辑;`,
                        `${indentToLevel(2)} * data 是需要上游返回的数据;`,
                        `${indentToLevel(2)} * 这样的 finally-then, 可以有 0 个, 或者, 若干个`,
                        `${indentToLevel(2)} */`,
                        `${indentToLevel()}});`
                    );
                    apiPathsSnippets[description] = {
                        scope,
                        prefix,
                        body,
                        description,
                    }
                }
            }

            return {
                ...apiPathsSnippets,
                ...entitiesSnippets,
            };
        })
        .then(snippets => {
            return jsonFormat(
                snippets,
                jsonIndentConfig
            )
        })
        .then(json => {
            function open() {
                return new Promise((resolve, reject) => {
                    opendir(snippetsDir, (error) => {
                        if (error) {
                            reject(error);
                        } else {
                            resolve();
                        }
                    });
                });
            }
            function make() {
                return new Promise((resolve, reject) => {
                    mkdir(snippetsDir, { recursive: true }, (error) => {
                        if (error) {
                            reject(error);
                        } else {
                            resolve();
                        }
                    });
                });
            }
            function write() {
                writeFile(
                    `${snippetsDir}/api-doc-${apiDocAlias}.code-snippets`,
                    json,
                    'utf8',
                    (error) => {
                        if (error) throw error;
                        console.log('done');
                    }
                );
            }
            open()
                .then(() => {
                    write();
                })
                .catch((error) => {
                    make()
                        .then(() => {
                            write();
                        })
                        .catch((error) => {
                            console.error(error);
                        });
                });
        });

    return void 0;
}
fetchAPIDoc();
